/***************************************************************************
                          system.cpp  -  models a hydraulic network
                             -------------------
    begin                : Sun Apr 9 2000
    copyright            : (C) 2000 by sdteffen
    email                : macke@heini.fbbwu.fh-lueneburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "system.h"

//libxml stuff
#include <parser.h>
#include <tree.h>

#include <stdlib.h>
#include <math.h>

//Constructor
ghSystem::ghSystem (const char *filename)
{
  //constants
  fMolarMass = 0.018;
  fNetworkRoughness = 0.00001;
  fRho = 1000.0;
  fKinematicViscosity = 1.10783094e-6;

  //arrays
  ghLines=NULL;
  ghNodes=NULL;
  fCoeff=NULL;

  xmlDocPtr
    xdpSystem = xmlParseFile (filename);
  if (xdpSystem == NULL) {
    printf ("XML Error");
    return;
  }

  //retrieve all nodes from XML data
  xmlNodePtr
    xnpXmlNode = xdpSystem->root->childs;

  while (xnpXmlNode != NULL) {
    printf ((const char *) xnpXmlNode->name);
    printf ("\n");
    if (!strcmp ((const char *) xnpXmlNode->name, "nodes"))
			parseNodes (xnpXmlNode);
    else if (!strcmp ((const char *) xnpXmlNode->name, "lines"))
			parseLines (xnpXmlNode);
    xnpXmlNode = xnpXmlNode->next;
  }
  xmlFreeDoc (xdpSystem);
};

void ghSystem::parseNodes (xmlNodePtr xnpXmlNodesNode)
{
  char *ccpString;
  int i;
  //Node data to be extracted from XML tree
  double fElevation;
  double fPressure;
  double fDemand;
  ghNodeType ntpType;

  iNumNodes = 0;
  xmlNodePtr xnpXmlNodesChild;

  xnpXmlNodesChild = xnpXmlNodesNode->childs;
  while(xnpXmlNodesChild != NULL) {
  	iNumNodes++;
    xnpXmlNodesChild = xnpXmlNodesChild->next;
  }
  ghNodes = (ghNode **)malloc(sizeof(ghNode)*iNumNodes);
  fCoeff = (double **)malloc(sizeof(double*)*iNumNodes);
  iEliminationOrder = (int *)malloc(sizeof(int)*iNumNodes);
  iNumCoeff = (int *)malloc(sizeof(int)*iNumNodes);
  if((ghNodes == NULL)||(fCoeff == NULL)||(iEliminationOrder == NULL)||
  		(iNumCoeff == NULL)){
  	printf("FATAL Error: Couldn't allocate memory.\n");
  	exit(0);
  }
  for(i=0;i<iNumNodes;i++){
  	fCoeff[i] = (double *)malloc(sizeof(double)*(iNumNodes+1));	
    if(fCoeff[i] == NULL){
	  	printf("FATAL Error: Couldn't allocate memory.\n");
  		exit(0);
	  }
  }
  iNumNodes=0;
  xnpXmlNodesChild = xnpXmlNodesNode->childs;
  while (xnpXmlNodesChild != NULL)
    {
      fElevation = 0.0;
      fPressure = 0.0;
      fDemand = 0.0;
      ntpType = GH_NODE_UNKNOWN_PRESSURE;

      ccpString =
	(char *) xmlGetProp (xnpXmlNodesChild, (xmlChar *) "elevation");
      if (ccpString != NULL)
	fElevation = atof (ccpString);
      ccpString =
	(char *) xmlGetProp (xnpXmlNodesChild, (xmlChar *) "pressure");
      if (ccpString != NULL)
	fPressure = atof (ccpString);
      ccpString = (char *) xmlGetProp (xnpXmlNodesChild, (xmlChar *) "type");
      if (ccpString != NULL)
	{
	  if (!strcmp (ccpString, "UnknownPressure"))
	    ntpType = GH_NODE_UNKNOWN_PRESSURE;
	  else if (!strcmp (ccpString, "ConstantPressure"))
	    ntpType = GH_NODE_CONSTANT_PRESSURE;
	  else if (!strcmp (ccpString, "RegulatePressure"))
	    ntpType = GH_NODE_REGULATE_PRESSURE;
	}
      ccpString =
	(char *) xmlGetProp (xnpXmlNodesChild, (xmlChar *) "demand");
      if (ccpString != NULL)
	fDemand = atof (ccpString);
      xnpXmlNodesChild = xnpXmlNodesChild->next;
      ghNodes[iNumNodes] =
	new ghNode (fPressure, fDemand, fElevation, ntpType);
      ghNodes[iNumNodes]->print ();
      iNumNodes++;
    }
};

void
ghSystem::parseLines (xmlNodePtr xnpXmlLinesNode)
{
  char*		    ccpString;
  xmlNodePtr  xnpXmlLinesChild;

  //Line data to be extracted from XML tree
  double 			fDiameter;
  double 			fRoughness;
  double 			fLength;
  int 				iFromNode;
  int 				iToNode;
  ghLineType 	ntpType;

  iNumLines = 0;

  xnpXmlLinesChild = xnpXmlLinesNode->childs;
  while(xnpXmlLinesChild != NULL) {
  	iNumLines++;
    xnpXmlLinesChild = xnpXmlLinesChild->next;
  }
  ghLines = (ghLine **)malloc(sizeof(ghLine)*iNumLines);
  if(ghLines == NULL){
  	printf("FATAL Error: Couldn't allocate memory.\n");
  	exit(0);
  }
  iNumLines=0;
  xnpXmlLinesChild = xnpXmlLinesNode->childs;


  while (xnpXmlLinesChild != NULL)
    {
      fDiameter = 0.0;
      fRoughness = 0.0;
      iFromNode = 0;
      iToNode = 0;
      ntpType = GH_LINE_LINE;

      ccpString =
	(char *) xmlGetProp (xnpXmlLinesChild, (xmlChar *) "diameter");
      if (ccpString != NULL)
	fDiameter = atof (ccpString);
      ccpString =
	(char *) xmlGetProp (xnpXmlLinesChild, (xmlChar *) "length");
      if (ccpString != NULL)
	fLength = atof (ccpString);
      ccpString =
	(char *) xmlGetProp (xnpXmlLinesChild, (xmlChar *) "roughness");
      if (ccpString != NULL)
	fRoughness = atof (ccpString);
      ccpString = (char *) xmlGetProp (xnpXmlLinesChild, (xmlChar *) "type");
      if (ccpString != NULL)
	{
	  if (!strcmp (ccpString, "Line"))
	    ntpType = GH_LINE_LINE;
	  else if (!strcmp (ccpString, "PumpForward"))
	    ntpType = GH_LINE_PUMP_FORWARD;
	  else if (!strcmp (ccpString, "PumpReverse"))
	    ntpType = GH_LINE_PUMP_REVERSE;
	  else if (!strcmp (ccpString, "PRVEnd"))
	    ntpType = GH_LINE_PRV_END;
	  else if (!strcmp (ccpString, "PRVStart"))
	    ntpType = GH_LINE_PRV_START;
	}
      ccpString = (char *) xmlGetProp (xnpXmlLinesChild, (xmlChar *) "from");
      if (ccpString != NULL)
	iFromNode = atoi (ccpString);
      ccpString = (char *) xmlGetProp (xnpXmlLinesChild, (xmlChar *) "to");
      if (ccpString != NULL)
	iToNode = atoi (ccpString);

      xnpXmlLinesChild = xnpXmlLinesChild->next;
      ghLines[iNumLines] =
				new ghLine (fDiameter, fLength, fRoughness, iFromNode, iToNode, ntpType);
      ghLines[iNumLines]->print ();
      iNumLines++;
    }
};

double
ghSystem::minimum (double value1, double value2)
{
  if (value1 > value2)
    return value2;
  else
    return value1;
};

double
ghSystem::maximum (double value1, double value2)
{
  if (value1 < value2)
    return value2;
  else
    return value1;
};

/** runs on iteration, returns the maximum Node Error */
double ghSystem::runIteration (int iNumIterations)
{
  int i, j, k, n;

  double  mfFlowv;	//temporary, TODO: remove/rename
  double  rtm;			//temporary, TODO: rename/remove

  /*Maximum Node Error, will be returned at the End of the method*/
  double fMaxNodeError = 0.0;

  double fAlpha; //kg/(s*Pa)

  double fAbsPressureDifference; // in Pa
  double fH2;//TODO: remove/rename
  double fH1;//TODO: remove/rename
  double fAA;//TODO: remove/rename
  double fPressureDifferenceVelocity; // in Pa
  double fPressureDifference; // in Pa
  /** TODO: find meaningful name */
  double fEdwlam;
  /** Epsilon */
  double *fPressureLossExponent;
  /*TODO: remove*/
  int iKr, iNr;
  /** TODO: make this local, find out meaning */
  double fRK;

	double *fNodeError;
  double *fStaticPressureDifference;

	fNodeError = (double *)malloc(sizeof(double)*iNumNodes);
	fStaticPressureDifference = (double *)malloc(sizeof(double)*iNumLines);
	fPressureLossExponent = (double *)malloc(sizeof(double)*iNumLines);
	if((fNodeError == NULL)||(fStaticPressureDifference == NULL)||
	   (fPressureLossExponent == NULL)){
  	printf("FATAL Error: Couldn't allocate memory.\n");
  	exit(0);
	}

  for (k = 1; k <= iNumNodes; ++k) {
    for (n = 1; n <= iNumNodes; ++n) {
		  fCoeff[k - 1][n - 1] = 0.0;
	  }
    if(ghNodes[k-1]->mType == GH_NODE_CONSTANT_PRESSURE) {
    	iKr = iEliminationOrder[k - 1];
      ghNodes[k-1]->mfDemand += fCoeff[iKr - 1][iNumNodes];
      fCoeff[k-1][k-1]=1.0;
    }
    if(ghNodes[k-1]->mType != GH_NODE_REGULATE_PRESSURE){
      fCoeff[k-1][k-1] -=	ghNodes[k-1]->mfLeakageRate /
							sqrt((fabs(ghNodes[k-1]->mfPressure) + 1.0) * 1e5);
		}
    fNodeError[k-1] =	ghNodes[k-1]->mfDemand
      									- ghNodes[k-1]->mfLeakageRate * sqrt (fabs(ghNodes[k-1]-> mfPressure)/1e5);
  }	//for loop over nodes, variable=k

  fMinPressureLossExponent = 2.0;
  fMaxPressureLossExponent = 1.0;

  for (i = 1; i <= iNumLines; i++) {
    k = ghLines[i-1]->miFromNode;
    iKr = iEliminationOrder[k-1];
    n = ghLines[i-1]->miToNode;
    iNr = iEliminationOrder[n-1];
    if ((ghLines[i-1]->mType == GH_LINE_LINE) ||
	      (ghLines[i-1]->mType == GH_LINE_PUMP_FORWARD) ||
	      (ghLines[i-1]->mType == GH_LINE_PUMP_REVERSE)) {
	    fMolarMass = 0.0;		//TODO: check this!!!
      if (fMolarMass > 0.0) {
      	//rtm = (fTemperature + (float)273.) * (float)8314. / fMolarMass;
      	rtm = 0;			//TODO: check rtm
      	fRho =((ghNodes[k-1]->mfPressure + ghNodes[n-1]->mfPressure) * 0.5 + fAmbientPressure) / rtm;
      }
      fStaticPressureDifference[i-1] =	(fRho -
					 fAmbientPressure / ((fAmbientTemperature + 273.0) * 8314.0) * 29.3) * 9.81 *(ghNodes[n - 1]->mfElevation -
					 ghNodes[k - 1]->mfElevation);
      if (ghLines[i-1]->mType == GH_LINE_PUMP_FORWARD) {
			  fStaticPressureDifference[i-1] -= ghLines[i-1]->mfDiameter;
			}
      else {
			  fStaticPressureDifference[i-1] += ghLines[i-1]->mfDiameter;
			}
      fPressureDifferenceVelocity = ghNodes[n-1]->mfPressure - ghNodes[k-1]->mfPressure +
																		fStaticPressureDifference[i-1];
      fPressureDifference =	fPressureDifferenceVelocity + fCoeff[iNr-1][iNumNodes] -
														fCoeff[iKr-1][iNumNodes];
      if (fabs(fPressureDifference) < 1.0)	{
	      fPressureDifference = 1.0;
	    }
      if (ghLines[i-1]->mType != GH_LINE_LINE) {
	    	fAlpha = fabs(ghLines[i-1]->mfLength / ghLines[i-1]->mfDiameter) *
      					pow (fabs( ghLines[i-1]->mfDiameter / fPressureDifference ), 1 - 1/fPressureLossExponent[i-1]);
    	}
      else {
        fAbsPressureDifference = fabs (fPressureDifference);
        if (ghLines[i-1]->mfLength < 0.0) {
          fAlpha =	ghLines[i-1]->mfDiameter * ghLines[i-1]->mfDiameter * 1.1107 /
    								sqrt (fabs (ghLines[i-1]->mfLength) * fAbsPressureDifference / fRho);
  			}
        else if (ghLines[i-1]->mfLength == 0)	{
	        fAlpha = fRho * ghLines[i-1]->mfDiameter * sqrt (fRho / (fAbsPressureDifference * 1e8));
  			}
        else {
          fH1 = ghLines[i-1]->mfDiameter * sqrt(ghLines[i-1]->mfDiameter * 2.0 *
    				       fRho * fAbsPressureDifference / ghLines[i-1]->mfLength);
          fAA = fH1 / (fRho * fKinematicViscosity);
          fRK = maximum (fabs(fNetworkRoughness), ghLines[i-1]->mfRoughness);
          fH2 = minimum (0.46,2.51 / fAA) + fRK / (ghLines[i-1]->mfDiameter * (float) 3.71);
          fH2 =2.51 / (maximum (2.6317, fAA) + fRK / (ghLines[i-1]->mfDiameter * (float) 3.71));
          fH2 = minimum (.95376, fH2);
          fEdwlam = minimum (log10(fH2)*(-2), fAA/64);
          fAlpha = ghLines[i-1]->mfDiameter * 0.7854 *	fH1 * fEdwlam / fAbsPressureDifference;
          mfFlowv = ghLines[i-1]->mfFlow;
          ghLines[i-1]->mfFlow = -fAlpha * fPressureDifference;
          if (iNumIterations <= 1) {
    			  goto FirstIterationLabel; //TODO remove this
    			}
          else {
            if (fabs(fPressureDifference - fPressureDifferenceVelocity) - 1.0 > 0.0)	{
              if (fabs(ghLines[i-1]->mfFlow - mfFlowv) - 0.001 > 0.0) {
                fPressureLossExponent[i-1] = -fAlpha*(fPressureDifference - fPressureDifferenceVelocity) /
                												(ghLines[i-1]->mfFlow - mfFlowv);
                fPressureLossExponent[i-1] = maximum(1, minimum(2,fPressureLossExponent[i-1]));
        			}
      			}
            fMinPressureLossExponent = minimum(fMinPressureLossExponent, fPressureLossExponent[i-1]);
            fMaxPressureLossExponent = maximum(fMaxPressureLossExponent, fPressureLossExponent[i-1]);
    			}
      	}
        fPressureLossExponent[i-1] = 2.0;
			}
      ghLines[i-1]->mfFlow = -fAlpha * fPressureDifference;
      fAlpha /= fPressureLossExponent[i-1];
		}
		else if (ghLines[i-1]->mType == GH_LINE_PRV_END) {
      ghLines[i-1]->mfFlow += fCoeff[iNr-1][iNumNodes];
	  }
	  else if (ghLines[i-1]->mType == GH_LINE_PRV_START) {
      ghLines[i-1]->mfFlow += fCoeff[iKr-1][iNumNodes];
	  }

    FirstIterationLabel:

    fNodeError[k-1] -= ghLines[i-1]->mfFlow;
    fNodeError[n-1] += ghLines[i-1]->mfFlow;
    if ((ghLines[i-1]->mType == GH_LINE_PRV_START)
		  || (ghLines[i-1]->mType == GH_LINE_PRV_END)) {
      if (ghNodes[k - 1]->mType == GH_NODE_REGULATE_PRESSURE)	{
        fCoeff[k-1][k-1] = (float) -1.;
        fCoeff[n-1][k-1] = (float) 1.;
			}
      if (ghNodes[n-1]->mType != GH_NODE_REGULATE_PRESSURE)	{
				continue;
			}
      else {
        fCoeff[k - 1][n - 1] = (float) -1.;
        fCoeff[n - 1][n - 1] = (float) 1.;
			}
		}
    else {
      if (ghNodes[n - 1]->mType == GH_NODE_UNKNOWN_PRESSURE) {
        fCoeff[k - 1][n - 1] = fAlpha;
        fCoeff[n - 1][n - 1] -= fAlpha;
        ++iNumCoeff[n - 1];
			}
      if (ghNodes[k - 1]->mType != GH_NODE_UNKNOWN_PRESSURE) {
      	continue;
			}
      else {
        fCoeff[k - 1][k - 1] -= fAlpha;
        ++iNumCoeff[k - 1];
        fCoeff[n - 1][k - 1] = fAlpha;
			  continue;
			}
		}
  }				//for loop over lines, variable i

  for (k = 1; k <= iNumNodes; ++k) {
    iKr = iEliminationOrder[k - 1];
    if (ghNodes[k-1]->mType == GH_NODE_UNKNOWN_PRESSURE) {
      ghNodes[k-1]->mfPressure += fCoeff[iKr - 1][iNumNodes];
    }
    fMaxNodeError = maximum (fMaxNodeError, fabs(fNodeError[iKr-1]));
    //possible error: iNumNodes and iKr might be switched for fCoeff
    fCoeff[iKr-1][iNumNodes] = (-1)*fNodeError[iKr-1];
  }				//end for loop over nodes, variable =k


  // virtual sorting of equations
  if (iNumIterations <= 0) {
    /* possible error next line: might be <=iNumNodes */
    for (n = 1; n < iNumNodes ; ++n) {
      for (k = 1; k <= iNumNodes-n; ++k) {
  		  if (iNumCoeff[k] - iNumCoeff[k-1] < 0) {
      	  i = iEliminationOrder[k];
      	  iEliminationOrder[k] = iEliminationOrder[k-1];
      	  iEliminationOrder[k-1] = i;
      	  j = iNumCoeff[k];
      	  iNumCoeff[k] = iNumCoeff[k-1];
      	  iNumCoeff[k-1] = j;
  	  	}
  		}			//end for loop over nodes, variable=k
    }				//end for loop over nodes, variable=n
  }
//  printCoeffMatrix();
  eliminateGauss();
//  printCoeffMatrix();
//  exit(0);

  //free allocated memory
  free(fNodeError);
  free(fStaticPressureDifference);
  free(fPressureLossExponent);

  return fMaxNodeError;
}

void ghSystem::eliminateGauss ()
{
  /* Local variables */
  static int i, j, k, n;
  static double z;
  static int k1, m1, ih, ir, jr, iKr, mr;

  n = iNumNodes + 1;
  m1 = iNumNodes - 1;

  for (k = 1; k <= iNumNodes-1; ++k) {
		iKr = iEliminationOrder[k - 1];
		k1 = k + 1;
		if (fCoeff[iKr-1][k-1] != (float)0.) {
	    goto L140;
		}
		else {
	  	  goto L110;
		}
L110:
		if (k1 - iNumNodes  <= 0) {
	  	  goto L112;
		}
		else {
	  	  goto L122;
		}
L112:
		for (i = k1; i <= iNumNodes; ++i) {
	    ir = iEliminationOrder[i - 1];
	    if (fCoeff[ir-1][k-1] != (float)0.) {
				goto L130;
	    }
	    else {
				goto L120;
	    }
L120:
	    ;
		} // end for loop variable = i
L122:
		printf("singular matrix k=%i\n",k1);
		exit(0);
L130:
		iKr = iEliminationOrder[i - 1];
		iEliminationOrder[i - 1] = iEliminationOrder[k - 1];
		iEliminationOrder[k - 1] = iKr;
L140:
		for (i = k1; i <= iNumNodes; ++i) {
	  	ir = iEliminationOrder[i - 1];
	  	z = fCoeff[ir-1][k-1] / fCoeff[iKr-1][k-1];	
    	for (j = k; j <= n; ++j) {
				fCoeff[ir-1][j-1] -= z * fCoeff[iKr-1][j-1];
	  	} // end for loop variable = j
		}  // end for loop variable = i
  } // end for loop variable = k
  mr = iEliminationOrder[iNumNodes-1];
  fCoeff[mr-1][n-1] /= fCoeff[mr-1][iNumNodes-1];
  for (ih = 2; ih <= iNumNodes; ++ih) {
		i = iNumNodes + 2 - ih;
		ir = iEliminationOrder[i - 2];
		for (j = i; j <= iNumNodes; ++j) {
	    	jr = iEliminationOrder[j - 1];
	  	  fCoeff[ir-1][n-1] -= fCoeff[ir-1][j-1] *fCoeff[jr-1][n-1];
		} // end for loop variable = j
		fCoeff[ir-1][n-1] /= fCoeff[ir-1][i-2];
	}	// end for loop variable = ih

}				/* method eliminateGauss */

/** analyses the Systems and returns the 
Number of Iterations needed. In case
of Error, 0 is returned */
int ghSystem::analyze ()
{
  int i, iNumIterations = 0;
  double fActualNodeError;

  //TODO: have C Array Boundaries
  for (i = 0; i < iNumNodes; i++) {
    iEliminationOrder[i] = i + 1;
    iNumCoeff[i] = 1;
    fCoeff[i][iNumNodes] = 0;
  }
  do {
    fActualNodeError = runIteration (iNumIterations);
    iNumIterations++;
  }
  while (fActualNodeError > fPermissibleNodeError);
  return iNumIterations;
}

/** prints Information about the System */
void
ghSystem::print ()
{
  int i;

  for (i = 0; i < iNumNodes; i++)
    ghNodes[i]->print ();
  for (i = 0; i < iNumLines; i++)
    ghLines[i]->print ();
}

/** Destructor, freeing the allocated arrays */
ghSystem::~ghSystem(){
	int i;
	if(ghNodes!=NULL){
		free(ghNodes);
	}
	if(ghLines!=NULL){
		free(ghLines);
	}
	if(fCoeff!=NULL){
		for(i=0;i<iNumNodes;i++){
			if(fCoeff[i]!=NULL){
				free(fCoeff[i]);
			}
		}
	}
	if(iNumCoeff != NULL){
		free(iNumCoeff);
	}
	if(iEliminationOrder != NULL) {
		free(iEliminationOrder);
	}
}


/** debug functions that prints the coefficient matrix */
void ghSystem::printCoeffMatrix(){
	int i,j;
	printf("***** Coeff Matrix ******\n");
	for(i=0;i<iNumNodes;i++){
		for(j=0;j<=iNumNodes;j++){
			printf("%f, ",fCoeff[i][j]);
		}
		printf("\n");
	}
	printf("*************************\n");
}




/** debug functions dumps the node error array */
void ghSystem::printNodeErrors(double *fNodeError){
	int i;
  printf("***** Node Errors ******\n");
	for(i=0;i<iNumNodes;i++){
		printf("%f, ",fNodeError[i]);
	}
	
	printf("\n*************************\n");
}



